import sys
import csv
import itertools
import urllib,urllib2
from collections import defaultdict


# take a list of query id and return a dictionary
def queryToUniprot(queryids, fromid, toid):
	queryresults = defaultdict(list)

	url = 'http://www.uniprot.org/mapping/'

	# find alternative uniprot accessions if query accessions are obsolete
	params = {
	'from':fromid,
	'to':toid,
	'format':'tab',
	'query':' '.join(queryids)
	}
	
	data = urllib.urlencode(params)
	request = urllib2.Request(url, data)
	contact = "" # Please set your email address here to help us debug in case of problems.
	request.add_header('User-Agent', 'Python %s' % contact)
	response = urllib2.urlopen(request)
	
	# skip the first line
	response.readline()
	for tokens in csv.reader(response, delimiter='\t'):
		queryresults[tokens[0]].append(tokens[1])
	
	return queryresults
	

# merge value lists from a dictionary
def collectValuesFromDict(adict):
	return list(set(itertools.chain.from_iterable(adict.values())))


# merge two dictionaries in given order
def mergeDict(dict1, dict2):
	results = defaultdict(list)
	for key1 in dict1.keys():
		dict1vals = dict1[key1]
		for key2 in dict1vals:
			results[key1] += dict2[key2]
	return results
	

def main():
	tags = ["MGI", "DFLAT", "GONE"]

	orthfile = "Output/Orth_GAF2.0."
	missinggsfile = "Output/MissingGS."
	tale = ".txt"

	# collect uniprot accessions that are missing gene symbols
	uniprot2genesymbol = {}
	for t in tags:
		inputfile = missinggsfile + t + tale
		
		with open(inputfile, 'r') as fi:
			for tokens in csv.reader(fi, delimiter='\t'):
				uniprot2genesymbol[tokens[1]] = []
	
	# query to uniprot after collecting all the missing uniprot accessions to reduce workload on their web server	
	# find uniprot accessions for collected uniprot accessions in case that they are obsolete
	u2u = queryToUniprot(uniprot2genesymbol.keys(), 'ACC', 'ACC')
	u2g = queryToUniprot(collectValuesFromDict(u2u), 'ACC', 'GENENAME')

	uniprot2genesymbol = mergeDict(u2u, u2g)
			
	# write missing gene symbols to orthfiles
	for t in tags:
		inputfile = missinggsfile + t + tale
		outputfile = orthfile + t + tale
		
		with open(inputfile, 'r') as fi, open(outputfile, 'a') as fo:
			for tokens in csv.reader(fi, delimiter = '\t'):
				uniprotacc = tokens[1]
				genesymbols = uniprot2genesymbol[uniprotacc]
				
				assert tokens[2] == ""
				
				for g in genesymbols:
					tokens[2] = g
					fo.write("%s\n" % '\t'.join(tokens))
						

if __name__ == "__main__":
    main()
