import sys
import csv
import itertools
import urllib,urllib2
from collections import defaultdict

from tempfile import mkstemp
from shutil import move
from os import remove, close


# update files with new namespace information
def replace(file_path, go2namespace):
	# create a temp file
	abs_path = file_path + ".temp"
	with open(abs_path, 'w') as new_file, open(file_path, 'r') as old_file:
		for line in old_file:
			tokens = line.strip().split('\t')
			goterm = tokens[4]
			if goterm in go2namespace:
				# the namespace should be unknown
				assert tokens[8] == 'X'
				tokens[8] = go2namespace[goterm]
				new_file.write("%s\n" % '\t'.join(tokens))
			else:
				new_file.write(line)
	# remove original file
	remove(file_path)
	# move new file
	move(abs_path, file_path)
				

# parse .obo line to get value
def getValue(linestr):
        ind = linestr.index(':')
        return linestr[ind+2:]


# merge value lists from a dictionary
def collectValuesFromDict(adict):
	return list(set(itertools.chain.from_iterable(adict.values())))


# merge two dictionaries in given order
def mergeDict(dict1, dict2):
	results = defaultdict(list)
	for key1 in dict1.keys():
		dict1vals = dict1[key1]
		for key2 in dict1vals:
			results[key1] += dict2[key2]
	return results
	

def main():
	tags = ["MGI", "DFLAT", "GONE"]

	obofile = "Input_Annotations/go.obo"
	orthfile = "Output/Orth_GAF2.0."
	missinggsfile = "Output/MissingPFC."
	tale = ".txt"

	# collect GO terms that are missing namespace information
	go2namespace = {}
	for t in tags:
		inputfile = missinggsfile + t + tale
		
		with open(inputfile, 'r') as fi:
			for line in fi:
				line = line.strip()
				if "GO:" in line:
					go2namespace[line] = ""
	gos = go2namespace.keys()

	# if there is no GO term missing namespace information, don't do the rest
	if len(gos) == 0:
		sys.exit()
	else:
		# parse .obo file again to find namespace information for terms in gos
		with open(obofile, 'r') as fi:
			newterm = False
			ids = []
			namespace = ""
			for line in fi:
				line = line.strip()
				if "[Term]" in line:
					newterm = True
				
				if newterm:
					if "id:" in line:
						val = getValue(line)
						if val in gos:
							ids.append(getValue(line))
					if "namespace:" in line:
						namespace = getValue(line)
						namespace = namespace[namespace.index('_')+1:][0].upper()
					if line == "":
						newterm = False					
	
						if len(ids) > 0:
							for g in ids:
								go2namespace[g] = namespace
							# check if the dictionary still has a key without a value
							try:
								ind = go2namespace.values().index("")
							except ValueError:
								# there is no need to parse .obo file any more
								# write the new information to files and quit the program
								break

		# write the new information to files and quit the program
		for t in tags:
			outputfile = orthfile + t + tale
			replace(outputfile, go2namespace)	


if __name__ == "__main__":
    main()
