#include <iostream>
#include <fstream>
#include <map>
#include <list>
#include <set>
#include <string>
#include <stdio.h>
using namespace std;

//Global variable: mapping of GO ID to GO term
map<string, string> goterm_map;

void createRelationshipMapping(string inFile, map<string, list<string> > &pmap,
                               map<string, list<string> > &cmap);

void printStringList(list <string> &mylist, map<string, string> &gomap);

void getAncestors(string term, list <string> &mylist, map<string,
					list<string> > &cmap);

void getAllChildren(string term, set <string> &myset, map<string,
					list<string> > &pmap);

void filterFile(string inFile, set <string> &childSet);

//The first argument is the gene ontology file
//The second argument is the file to be filtered
int main(int argc, char* argv[])
{
	//GO mapping: a parent term => a list of children terms
	map<string, list<string> > parent_child_map;

	//GO mapping: a child term => a list of parent terms
	map<string, list<string> > child_parent_map;

	if(argc == 3)
	{
		// remove this part to automate this process
		// instead of taking user input, assign value to "term"
		//string term;
		//cout << "Enter the GO ID to filter by: ";
		//getline(cin, term);

		string term = "GO:0032502";

		//Create mappings of GO terms
		//taking only the "is_a" and "part_of" relationships only
		createRelationshipMapping(argv[1], parent_child_map, child_parent_map);
		set<string> children;
		getAllChildren(term, children, parent_child_map);
		children.insert(term);
		filterFile(argv[2], children);
		return 0;
    }
	else
		cout << "Arguments not passed correctly." << endl;
	return 0;
}

//helper function: given a GO ID, get the list of ancestors
void getAncestors(string term, list <string> &mylist,
					map<string, list<string> > &cmap)
{
	//Get the list of immediate parents
	list <string> temp = cmap[term];

	//if no parents, then the child must be a root value
	if(temp.size() == 0)
		return;
	else
    {
		//for every parent, add to the ancestor list,
		//then their ancestors recursively
		list<string>::iterator it;
		for(it = temp.begin(); it != temp.end(); it++)
		{
			string parent = (*it);
			mylist.push_back(parent);
			getAncestors(parent, mylist, cmap);
		}
	}
}

//helper function: given a GO term, get the set of all children
void getAllChildren(string term, set <string> &myset,
					map<string, list<string> > &pmap)
{
	//Get the list of immediate children
	list <string> temp = pmap[term];

	//if no children, then the child must be a leaf
	if(temp.size() == 0)
		return;
	else
    {
		//for every child, add to the children set,
		//then their children recursively
		list<string>::iterator it;
		for(it = temp.begin(); it != temp.end(); it++)
		{
			string child = (*it);
			myset.insert(child);
			getAllChildren(child, myset, pmap);
		}
	}
}

void filterFile(string inFile, set <string> &childSet)
{
	//open the input file
	ifstream fin;
	fin.open(inFile.c_str());

	//open the output file
	ofstream fout;
	fout.open("Output/filtered.txt");

	if(fin.is_open())
	{
		string goTerm;
		string db, db_obj_id, db_obj_sym, qual;
		set<string>::iterator temp;
		string restOfLine;

		while (!fin.eof())
		{
			//super annoying but whatever
			getline(fin, db, '\t');
			getline(fin, db_obj_id, '\t');
			getline(fin, db_obj_sym, '\t');
			getline(fin, qual, '\t');
			getline(fin, goTerm, '\t');
			getline(fin, restOfLine);

			//check if the goTerm is in the set
			temp = childSet.find(goTerm);
			if(temp != childSet.end())
			{
				//print the line to the filtered file
				//fout << goTerm << '\t' << restOfLine << '\n';
				fout << db << '\t' << db_obj_id << "\t" << db_obj_sym << '\t' << qual << "\t" << goTerm << '\t' << restOfLine << '\n';
			}
		}
	}
	else
		cout << "Cannot open the input file!" <<endl;
	fin.close();
	fout.close();
}

//helper function: really bad name for this function but
//given a list of GO ids, print their GO terms
 void printStringList(list <string> &mylist, map<string, string> &gomap)
 {
   list<string>::iterator it;
   for(it = mylist.begin(); it != mylist.end(); it++)
     {
       string id = (*it);
       string name = gomap[id];
       cout<<id<<" '"<<name<<"'"<<endl;
     }
 }

//Create a relational mapping of GO terms => child GO terms
//Takes the gene ontology file and extracts relationships such as
//"is_a" and "part_of": please refer to the format of the GO file
void createRelationshipMapping(string inFile,
			map<string, list<string> > &pmap,
			map<string, list<string> > &cmap)
{
	ifstream fin;
	fin.open(inFile.c_str());

	if(fin.is_open())
	{
		string line;
		string child_name;
		string child_id;
		string parent_term;
		string parent_id;
		int pos;

		while(!fin.eof())
		{
			getline(fin, line);

			//If the line that is read is a GO ID,
			//then extract the GO term (next line of the file)
			if(line.find("id:") == 0)
			{
				//Extract GO id:
				pos = line.find("GO:");
				if(pos > 0)
				{
					child_id = line.substr(pos);
				}

				//Go to the next line and grab the GO term
				getline(fin, line);
				pos = line.find("name:");
				if(pos == 0) {
					child_name = line.substr(pos+6);
					//cout << "Child name is: " << child_name <<endl;
				}

				//Now search the next few lines for relationships
				//until you hit another GO term or you reach
				//the end of the file
				while(line.find("[Term]") == -1 && !fin.eof())
				{
					getline(fin, line);

					//For every line, see if there is a "is_a"
					//or "part_of" relationship mentioned.
					//If they are mentioned, extract the info
					//and put it in the mapping structure

					pos = line.find("is_a:");
					if(pos == 0)
					{
						int pos = line.find("GO:");
						if(pos != -1)
						{
							parent_id = line.substr(pos, 10);
							pos = line.find("!");
							parent_term = line.substr(pos+2);
							//cout << "The parent term is: " << parent_term << endl;
							/*
							pmap[parent_term].push_back(child_name);
							cmap[child_name].push_back(parent_term);
							*/

							pmap[parent_id].push_back(child_id);
							cmap[child_id].push_back(parent_id);
						}
					}

					pos = line.find("part_of");
					if(pos >= 0)
					{
						pos = line.find("GO:");
						if(pos != -1)
						{
							parent_id = line.substr(pos, 10);
							pos = line.find("!");
							parent_term = line.substr(pos+2);
							//cout << "The parent term is: " << parent_term << endl;
							/*
							pmap[parent_term].push_back(child_name);
							cmap[child_name].push_back(parent_term);
							*/

							pmap[parent_id].push_back(child_id);
							cmap[child_id].push_back(parent_id);
						}
					}
				}
			}
		}
	}
	else
		cout<<"Could not open input gene ontology file"<<endl;
	fin.close();
}
