#include <iostream>
#include <fstream>
#include <map>
#include <list>
#include <string>
#include <stdio.h>
using namespace std;

//Global variable: mapping of GO ID to ontology (F/P/C)
map<string, string> domain;

//Global variable: mapping of GO ID to GO term
map<string, string> goterm_map;

//Global variable: mapping of GO id to depth
map<string, int> depth;

//Global variable: mapping of alt_id => real_id
map<string, string> altid;

void assignDepth(string s, int d, map<string, list<string> > &tree)
{
  int num = depth[s];
  if(num == 0)
    depth[s] = d;
  if(num < d)
    depth[s] = d;

  list<string> temp = tree[s];
  if(temp.size() == 0)
    return;

  list<string>::iterator it;
  for(it = temp.begin(); it != temp.end(); it++)
    {
      assignDepth(*it, d+1, tree);
    }
}


void createRelationshipMapping(string inFile, map<string, list<string> > &pmap,
                               map<string, list<string> > &cmap);

void printStringList(list <string> &mylist, map<string, string> &gomap);

void getAncestors(string term, list <string> &mylist, map<string,
		  list<string> > &pmap);

//input: GOA file to be exploded
//output: dummy file to hold exploded entries
//pmap: mapping from children to their parents in GO tree
//This function maps terms up the GO tree. The logic behind this
//function is: whatever node in the ontology tree a gene belongs
//to, it should also belong to all of its ancestors in the tree.
void explode(string input, string output,  map<string, list<string> > &pmap)
{
  //open the input file
  ifstream fin;
  fin.open(input.c_str());

  //open the output file
  ofstream fout;
  fout.open(output.c_str());

  string tempGene;
  string tempGOterm;
  string tempSet;
  string tempEvidenceCode;
  string tempQualifier;
  string temp;
  string garbage;
  bool firstLine = true;

  int count = 1;

  //the dictionary data structure
  map <string, list<string> > ancestors;

  if(fin.is_open())
    {
      //PART 1 (of 2) of the function reads in the input file:
      //for each line of the input GOA file
      while (!fin.eof())
	{
	  getline(fin, temp, '\t');
	  //extract the gene (column 3)
	  if (count == 3)
	    {
	      tempGene = temp;
	    }
	  //extract Qualifier (column 4)
	  if (count == 4)
	    tempQualifier = temp;

	  //extract GO term (column 5)
	  if (count == 5)
	    {
	      tempGOterm = temp;
	    }

	  //extract Evidence Code (column 7)
	  if (count == 7)
	    tempEvidenceCode = temp;

	  //extract 'F'/'P'/'C' (column 9, indicating function, process, or
	  //cell component)
	  if (count == 9)
	    {
	      tempSet = temp;
	    }

	  count++;

	  //after column 9, all relevant information has been parsed
	  //and writing out can begin
	  if (count > 9)
	    {
	      //the rest of the line is garbage
	      getline(fin, garbage);
	      //reset count for next iteration
	      count = 1;
	      //4 criteria for inclusion in final data
	      //  (1) gene is not an exact repeated entry
	      //  (2) gene has a valid evidence code (IDA, IPI, IMP,
	      //      IGI, IEP, ISS, TAS)
	      //  (3) gene does not have the qualifier 'NOT'
	      //  (4) gene does not have too broad a GO term
	      //      (0008150, 0003674 and 0005575 respond to
	      //       molecular function, biological process
	      //       and cellular process, which are TOO BROAD)
	      if ((tempEvidenceCode == "IDA"
		   || tempEvidenceCode == "IPI" || tempEvidenceCode ==
		   "IMP" || tempEvidenceCode == "IGI" || tempEvidenceCode
		   == "IEP" || tempEvidenceCode == "ISS" ||
		   tempEvidenceCode == "TAS") && tempQualifier != "NOT"
		  && tempGOterm != "GO:0008150" && tempGOterm !=
		  "GO:0003674" && tempGOterm != "GO:0005575")
		{
		  //write out the information that will be needed later
		  //for creating the GMT files
		  if (firstLine)
		    {
		      fout << tempGOterm << '\t' << tempSet << '\t' <<
			tempGene  << '\t' << "xx" << '\t' << "xx";
		      firstLine = false;
		    }
		  else
		    fout << '\n' << tempGOterm << '\t' << tempSet << '\t'
			 << tempGene << '\t' << "xx" << '\t' << "xx";

		  list <string> temp;
		  //PART 2 of the function does the explosion

		  //check if current gene is in the dictionary:
		  //if not, get the ancestors and add
		  //them in the data structure
		  temp = ancestors[tempGOterm];
		  if(temp.empty())
		    {
		      getAncestors(tempGOterm, temp, pmap);
		    }

		  list<string>::iterator it;
		  for(it = temp.begin(); it!=temp.end(); it++)
		    {
		      string anc = *it;

		      //prints out GO ID to ontology: commented out
		      //if(domain[anc] != domain[tempGOterm])
		      //cout<<tempGOterm<<" different domain from "<<anc<<endl;

		      //assign the current GO term to each of its ancestors
		      //unless ancestor is one of the 3 roots of the tree
		      if ( anc != "GO:0008150" && anc !=
			   "GO:0003674" && anc != "GO:0005575")
			fout << '\n' << anc << '\t' << tempSet << '\t'
			     << tempGene << '\t' << "xx" << '\t' << "xx";
		    }
		}
	    }
	}
    }
  else
    cout<<"Could not open human GOA file"<<endl;
  fin.close();
  fout.close();
}

//input: file name
//output: file name
//Reads the exploded input file, sorts it by GO ID, and writes
//out the sorted data to the output file
void sortByGO(string input, string output)
{
  ifstream fin;

  //open the input file to be read
  fin.open(input.c_str());

   //open the output file
   ofstream fout;
   fout.open(output.c_str());

   //create a list and temp variables to hold gene data
   list<string> mylist;
   string line;
   string preline;

   if (fin.is_open() && fout.is_open())
     {
       while (!fin.eof())
	 {
	   //treat a line of gene data as a string
	   getline(fin, line);
	   //discard errant blank lines
	   if (!line.empty())
	     mylist.push_back(line);
	 }

       //exploded file has GO terms at the front of each line,
       //so built-in sort() will work correctly
       mylist.sort();

       while (!mylist.empty())
	 {
	   //rewrite now sorted genes list to output file
	   line = mylist.front();
	   mylist.pop_front();
	   //check if duplicate gene are inherited
	   //due to the explosion. If it is a
	   //duplicate, don't print it to the file
	   if (line != preline)
	     {
	       fout << line << '\n';
	       preline = line;
	     }
	 }
     }
   else
     cout << "Couldn't open file" << endl;

   fin.close();
   fout.close();
 }

//Takes a GMT file and converts all GO IDs to their
//respective GO terms
void convertGOID(string input, string output_GMT)
{
  ifstream fin;
  fin.open(input.c_str());

  ofstream fout;
  fout.open(output_GMT.c_str());

  string goterm;
  string line;
  bool firstLine = true;

  if (fin.is_open() && fout.is_open())
    {
      while (!fin.eof())
        {
	  //assuming first tab has GO ID
          getline(fin, line, '\t');
          goterm = goterm_map[line];

	  //get the rest of the line
	  getline(fin, line);

	  //if the GO term exists, then print it out
	  //with the rest of the line
	  if(!goterm.empty())
	    {
	      if(firstLine)
		{
		  fout<<goterm<<'\t'<<line;
		  firstLine = false;
		}
	      else
		{
		  fout<<'\n'<<goterm<<'\t'<<line;
		}
	    }
	}
    }
}

//input: file name
//output: file name
//removes gene sets of size <= 1 from the input file
//and writes adequately sized sets to the output file in GMT format
void checkSize(string input, string output_GMT)
{
  ifstream fin;
  fin.open(input.c_str());

  ofstream fout;
  fout.open(output_GMT.c_str());

  string line;
  bool firstLine = true;

  if (fin.is_open() && fout.is_open())
    {
      while (!fin.eof())
        {
	  //size of the set is determined by counting tabs
          getline(fin, line);
          int countTab = 0;
          int checkTab = 0;

          while (checkTab >= 0)
            {
              checkTab = line.find('\t', checkTab);
              countTab++;

              if (checkTab != -1)
                checkTab++;
            }

          //given tab delimited string, lines with fewer than 2 tabs
          //should be omitted as sets that are too small
          if (countTab >= 2)
	    {
	    if(firstLine)
	      {
		firstLine = false;
		fout << line;
		cout << line;
	      }
	    else
	      {
		fout<<'\n'<<line;
	      }
	    }
        }

      fin.close();
      fout.close();
    }

  return;
}


//function parse()
//input: inputGOA      (string, exploded sorted file derived from GO database)
//       outputF_GMT   (string, file that will be created to store
//                      molecular function genes in GMT format)
//       outputP_GMT   (string, file that will be created to store
//                      biological process genes in GMT format)
//       outputC_GMT   (string, file that will be created to store
//                      cellular component genes in GMT format)
//reads in the input file in the format created by the explosion function
//and outputs the GMT files
void parse(string inputGOA, string outputF_GMT, string outputP_GMT,
	   string outputC_GMT, string outputFPC_GMT)
{
  //genes sorted by GO term
  sortByGO(inputGOA, "dummy2.txt");

  //open file streams
  ifstream fin;
  fin.open("dummy2.txt");

  ofstream foutF;
  foutF.open("dummy_f.txt");
  ofstream foutP;
  foutP.open("dummy_p.txt");
  ofstream foutC;
  foutC.open("dummy_c.txt");


  //temp Variables
  string tempGene = "";
  string tempGOterm = "";
  string tempSet;
  string temp;
  string garbage;

  bool firstLineF = true;
  bool firstLineP = true;
  bool firstLineC = true;

  bool sameSet = false;
  bool sameGO = false;

  if (fin.is_open() && foutF.is_open() && foutP.is_open() &&
      foutC.is_open())
    {
      int count = 1;
      //loop through all of the gene entries
      while (!fin.eof())
        {
	  getline(fin, temp, '\t');
	  //extract GO term (column 1)
	  if(count == 1)
	    {
	      //check if term is the same as the previous one
	      if(tempGOterm == temp)
		sameGO = true;
	      else
		{
		  tempGOterm = temp;
		  sameGO = false;
		}
	    }
	  //extract F/P/C (column 2)
	  if(count == 2)
	    {
	      if(tempSet == temp)
		sameSet = true;
	      else
		{
		  tempSet = temp;
		  sameSet = false;
		}
	    }
	  //extract gene ID (column 3)
	  if(count == 3)
	    {
	      tempGene = temp;
	    }
	  count++;
	  //after the first three columns, all relavent info has been extracted
	  if(count > 3)
	    {
	      getline(fin, garbage);
	      count = 1;

	      tempSet = domain[tempGOterm];
	      //write out genes to separate files based on F/P/C
	      if(tempSet == "F")
		{
		  if(firstLineF == true)
		    {
		      foutF<<tempGOterm<<'\t'<<depth[tempGOterm]<<'\t'<<tempGene;
		      firstLineF = false;
		    }
		  else
		    {
		      if(sameGO)
			{
			  foutF<<'\t'<<tempGene;
			}
		      else
			{
			  foutF<<'\n'<<tempGOterm<<'\t'<<depth[tempGOterm]<<'\t'<<tempGene;
			}
		    }
		}

	      if(tempSet == "P")
		{
		  if(firstLineP == true)
		    {
		      foutP<<tempGOterm<<'\t'<<depth[tempGOterm]<<'\t'<<tempGene;
                      firstLineP = false;
		    }
		  else
		    {
		      if(sameGO)
                        {
                          foutP<<'\t'<<tempGene;
                        }
                      else
                        {
                          foutP<<'\n'<<tempGOterm<<'\t'<<depth[tempGOterm]<<
			    '\t'<<tempGene;
                        }
		    }
		}

	      if(tempSet == "C")
		{
		  if(firstLineC == true)
		    {
		      foutC<<tempGOterm<<'\t'<<depth[tempGOterm]<<'\t'<<tempGene;
		      firstLineC = false;
		    }
		  else
		    {
		      if(sameGO)
                        {
                          foutC<<'\t'<<tempGene;
                        }
                      else
                        {
                          foutC<<'\n'<<tempGOterm<<'\t'<<depth[tempGOterm]<<
			    '\t'<<tempGene;
                        }
		    }
		}

	    }
	}//end of while loop
    }
  //close file streams
  fin.close();
  foutF.close();
  foutP.close();
  foutC.close();

  //remove small gene sets from each file
  checkSize("dummy_f.txt", "dummy2_f.txt");
  checkSize("dummy_p.txt", "dummy2_p.txt");
  checkSize("dummy_c.txt", "dummy2_c.txt");

  //convert GO IDs into GO term names
  convertGOID("dummy2_f.txt", outputF_GMT);
  convertGOID("dummy2_p.txt", outputP_GMT);
  convertGOID("dummy2_c.txt", outputC_GMT);

  //create composite file
  foutF.open(outputFPC_GMT.c_str());
  string line;

  fin.open(outputF_GMT.c_str());
  while (!fin.eof())
    {
      getline(fin, line);
      foutF << line << '\n';
    }
  fin.close();
  fin.open(outputP_GMT.c_str());
  while (!fin.eof())
    {
      getline(fin, line);
      foutF << line << '\n';
    }
  fin.close();
  fin.open(outputC_GMT.c_str());
  while (!fin.eof())
    {
      getline(fin, line);
      foutF << line << '\n';
    }
  fin.close();
  foutF.close();

  //Remove dummy files: commented out for now
  /*
  //remove temporary files
  if(remove("dummy.txt") != 0)
  cout<<"Error in deleting the file dummy.txt"<<endl;
  if(remove("dummy2.txt") != 0)
  cout<<"Error in deleting the file dummy.txt"<<endl;
  if(remove("dummy_f.txt") != 0)
  cout<<"Error in deleting the file dummy_f.txt"<<endl;
  if(remove("dummy_p.txt") != 0)
  cout<<"Error in deleting the file dummy_p.txt"<<endl;
  if(remove("dummy_c.txt") != 0)
  cout<<"Error in deleting the file dummy_c.txt"<<endl;
  if(remove("dummy2_f.txt") != 0)
  cout<<"Error in deleting the file dummy2_f.txt"<<endl;
  if(remove("dummy2_p.txt") != 0)
  cout<<"Error in deleting the file dummy2_p.txt"<<endl;
  if(remove("dummy2_c.txt") != 0)
  cout<<"Error in deleting the file dummy2_c.txt"<<endl;
  if(remove("clean.txt") != 0)
  cout<<"Error in deleting the file dummy2_c.txt"<<endl;
  */
}

void replace_alt_ids(string inFile, string outFile)
{
  ifstream fin;
  fin.open(inFile.c_str());

  ofstream fout;
  fout.open(outFile.c_str());

  if(fin.is_open())
    {
      string line;
      int pos;

      while(!fin.eof())
        {
          getline(fin, line);
          pos = line.find("GO:");
          if(pos != -1)
	    {
              string id = line.substr(pos, 10);
              string real_id = altid[id];
              if(!real_id.empty())
                {
                  line.replace(pos,10,real_id);
                }
            }
          fout << line <<endl;
        }
    }
}



//The first argument is the GOA human input file.goa
//The second argument is the F.gmt file to be created
//The third argument is the P.gmt file to be created
//The fourth argument is the C.gmt file to be created
//The fifth argument is the gene ontology file
//The sixth argument is the FPC.gmt file to be created
int main(int argc, char* argv[])
{
  //GO mapping: a parent term => a list of children terms
  map<string, list<string> > parent_child_map;

  //GO mapping: a child term => a list of parent terms
  map<string, list<string> > child_parent_map;

  if(argc == 7)
    {
      //Create mappings of GO terms with parents
      //taking only the "is_a" and "part_of" relationships only
      createRelationshipMapping(argv[5], parent_child_map, child_parent_map);
      replace_alt_ids(argv[1], "clean.txt");

      //map all genes up the GO tree and rewrite to dummy.txt
      explode("clean.txt", "dummy.txt", child_parent_map);

      //parse the dummy.txt file and create the output files
      parse("dummy.txt", argv[2], argv[3], argv[4], argv[6]);

    }
  else
    cout << "Arguments not passed correctly." << endl;

  return 0;
}

//helper function: given a GO ID, get the list of ancestors
 void getAncestors(string term, list <string> &mylist,
		   map<string, list<string> > &pmap)
 {
   //Get the list of immediate parents
   list <string> temp = pmap[term];

   //if no parents, then the child must be a root value
   if(temp.size() == 0)
     return;
   else
     {
       //for every parent, add to the ancestor list,
       //then their ancestors recursively
       list<string>::iterator it;
       for(it = temp.begin(); it != temp.end(); it++)
	 {
	   string parent = (*it);
	   mylist.push_back(parent);
	   getAncestors(parent, mylist, pmap);
	 }
     }
 }

//helper function: really bad name for this function but
//given a list of GO ids, print their GO terms
 void printStringList(list <string> &mylist, map<string, string> &gomap)
 {
   list<string>::iterator it;
   for(it = mylist.begin(); it != mylist.end(); it++)
     {
       string id = (*it);
       string name = gomap[id];
       cout<<id<<" '"<<name<<"'"<<endl;
     }
 }

//Create a relational mapping of GO IDs => ancestor GO IDs
//NOTE: the mapping of parent GO IDs => child GO IDs is omitted!
//Takes the gene ontology file and extracts relationships such as
//"is_a" and "part_of": please refer to the format of the GO file
 void createRelationshipMapping(string inFile,
				map<string, list<string> > &pmap,
				map<string, list<string> > &cmap)
 {
   //strings used for assigning depths
  string mole_fxn = "GO:0003674";
  string bio_proc = "GO:0008150";
  string cell_comp = "GO:0005575";

   ifstream fin;
   fin.open(inFile.c_str());

   if(fin.is_open())
     {
       string line;
       int pos;
       while(!fin.eof())
	 {
	   getline(fin, line);

	   //If the line that is read is a GO ID,
	   //then extract the GO term (next line of the file)
	   //and then the GO id's namespace (next line after)
	   if(line.find("id:") == 0)
	     {
	       string child_id;
	       string child_name;
	       pos = line.find("GO:");
	       if(pos > 0)
		 child_id = line.substr(pos);

	       //Grab the next line, assuming it's the GO term
	       getline(fin, line);
	       pos = line.find("name:");
	       if(pos == 0)
		 child_name = line.substr(pos+6);
	       goterm_map[child_id] = child_name;

	       //Grab the next line, assuming it's the namespace
	       getline(fin, line);
	       string name_space = line.substr(11);
	       if(name_space == "biological_process")
		 domain[child_id] = "P";
	       if(name_space == "molecular_function")
		 domain[child_id] = "F";
	       if(name_space == "cellular_component")
		 domain[child_id] = "C";

	       //Now search the next few lines for relationships
	       //until you hit another GO term or you reach
	       //the end of the file
	       while(line.find("[Term]") == -1 && !fin.eof())
		 {
		   string parent_id;
		   getline(fin, line);

		   //For every line, see if there is a "is_a"
		   //or "part_of" relationship mentioned.
		   //If they are mentioned, extract the info
		   //and put it in the mapping structure

		   pos = line.find("is_a:");
		   if(pos == 0)
		     {
		       int pos = line.find("GO:");
		       if(pos != -1)
			 {
			   parent_id = line.substr(pos, 10);
			   pmap[parent_id].push_back(child_id);
			   cmap[child_id].push_back(parent_id);
			 }
		     }

		   pos = line.find("part_of");
		   if(pos >= 0)
		     {
		       pos = line.find("GO:");
		       if(pos != -1)
			 {
			   parent_id = line.substr(pos, 10);
			   pmap[parent_id].push_back(child_id);
			   cmap[child_id].push_back(parent_id);
			 }
		     }

		   pos = line.find("alt_id:");
		   if(pos == 0)
		     {
		       pos = line.find("GO:");
		       if(pos != -1)
			 {
			   string alt_id = line.substr(pos, 10);
			   altid[alt_id] = child_id;
			 }
		     }
		 }
	     }
	 }
       assignDepth(mole_fxn, 1, pmap);
       assignDepth(bio_proc, 1, pmap);
       assignDepth(cell_comp, 1, pmap);
     }
   else
     cout<<"Could not open input gene ontology file"<<endl;
   fin.close();
 }
