import sys
import datetime

# argument list
# 1) go.obo file
# 2) output file

# from a line, get substring that contains value
def getValue(linestr):
	ind = linestr.index(':')
	return linestr[ind+2:]


def main():
	gofile = sys.argv[1]
	outputfile = sys.argv[2]
	
	# load go.obo file
	allids = []
	allaltids = []
	allaltnamespace = {}

	newterm = False
	id = ""
	name = ""
	namespace = ""
	altids = []
	obsolete = ""
	
	# get alternative ids and output them to a file at the same time
	with open(gofile, 'r') as fi, open(outputfile, 'w') as fo:
	
		# write comments to the output file
		fo.write("! date generated: %s\n" % datetime.datetime.now())
		fo.write("! generated from file %s\n" % gofile)
		fo.write("! \n")
		fo.write("! GO IDs (primary and secondary) and text strings\n")
		fo.write("! GO:0000000 (primary) [tab] GO:0000000 (secondary, separated by space(s) if >1) [tab] text string [tab] F|P|C\n")
		fo.write("! where F = molecular function, P = biological process, C = cellular component\n")
		fo.write("! obs = term is obsolete\n! \n")
		
		# load .obo file and collect information about each GO term
		for line in fi:
			line = line.strip()
			
			if "[Term]" in line:
				newterm = True

			if newterm:
				if line[:3] == "id:":
					id = getValue(line)
					allids.append(id)
				if line[:7] == "alt_id:":
					altids.append(getValue(line))
					allaltids += altids
				if "name:" in line:
					name = getValue(line)
				if "namespace:" in line:
					namespace = getValue(line)
					namespace = namespace[namespace.index('_')+1:][0].upper()					
				if "is_obsolete:" in line:
					if getValue(line) == "true":
						obsolete = "obs"
				
				if line == "":				
					# end of one term
					fo.write("%s\t%s\t%s\t%s\t%s\n" % (id, " ".join(altids), name, namespace, obsolete))
										
					if len(altids) > 0:
						for a in altids:
							allaltnamespace[a] = namespace

					id = ""
					name = ""
					namespace = ""
					altids = []
					obsolete = ""					
					newterm = False

		# find a new term and add to a file
		# definition of a new term: only appeared as alt_id without name or namespace of its own
		newids = list(set(allaltids) - set(allids))
		fo.write("! new ones added that serve only as alternative ids\n")
		for n in newids:
			fo.write("%s\t%s\t%s\t%s\t%s\n" % (n, "", "", allaltnamespace[n], ""))

if __name__ == "__main__":
    main()
